from MI_Pack.KEGG import Compound
from MI_Pack.KEGG import Map
from MI_Pack.SqliteManager import *
from MI_Pack import Formula

"""
# Copyright 2001 by Tarjei Mikkelsen.  All rights reserved.
# This code is part of the Biopython distribution and governed by its
# license.  Please see the LICENSE file that should have been included
# as part of this package.
# This module provides code to work with data from the KEGG database.

References:

Kanehisa, M. and Goto, S.; KEGG: Kyoto Encyclopedia of Genes and Genomes.
Nucleic Acids Res. 28, 29-34 (2000). 

URL: http://www.genome.ad.jp/kegg/
"""

KEGG_ITEM_LENGTH = 12
KEGG_LINE_LENGTH = 80
KEGG_DATA_LENGTH = KEGG_LINE_LENGTH - KEGG_ITEM_LENGTH

_default_wrap = lambda indent: [indent, "", (" ", "", 1, 0)]

def _wrap_kegg(line, max_width = KEGG_DATA_LENGTH, wrap_rule = _default_wrap):
    """Wraps the input line  for KEGG output.

    Arguments:
    
    o info - String holding the information we want wrapped
    for KEGG output.
    o max_width - Maximum width of a line.
    o wrap_rule - A wrap rule (see above) for deciding how to split
    strings that must be wrapped.
    """
    s = ""
    wrapped_line = ""
    indent =  " " * wrap_rule[0]
    connect = wrap_rule[1]
    rules =   wrap_rule[2:]
    while 1:
        if len(line) <= max_width:
            wrapped_line = wrapped_line + line
            s = s + wrapped_line
            break
        else:
            did_split = 0
            for rule in rules:
                to = max_width
                if not rule[2]:
                    to = to + len(rule[0])
                split_idx = line.rfind(rule[0], 0, to)
                if split_idx > -1:
                    if rule[2] and rule[3]:
                        split_idx = split_idx + len(rule[0])
                    wrapped_line = wrapped_line + line[0:split_idx] + "\n"
                    if not rule[3]:
                        split_idx = split_idx + len(rule[0])
                    line = indent + rule[1] + line[split_idx:]
                    did_split = 1
                    break
            if not did_split:
                wrapped_line = wrapped_line + line[0:max_width] + "\n"
                line = indent + connect + line[max_width:]
    return s


def _write_kegg(item, info, indent = KEGG_ITEM_LENGTH):
    """Write a indented KEGG record item.
    
    Arguments:
    
    o item - The name of the item to be written.
    o info - The (wrapped) information to write.
    o indent - Width of item field.
    """
    s = ""
    for line in info:
        partial_lines = line.splitlines()
        for l in partial_lines:
            s = s + item.ljust(indent) + l + "\n"
            if item is not "":  # ensure item is only written on first line
                item = ""
    return s

"""
#####################################
# END CODE
#####################################
"""


"""
#####################################
# Ralf Weber 2009
# This code is part of MI-Pack
# This module provides code to work with data from the KEGG database.
#
# http://www.biosciences.bham.ac.uk/labs/viant/
#####################################
"""

def Compounds(file_compounds, atoms, ions, ion_mode, debug=0):

    database = SQLiteManager("MI_DB_%s" % ion_mode)
    database.create_table("compounds")
    
    print
    print "Parsing compounds (file: %s) to MI_DB_%s.sqlite" % (file_compounds, ion_mode)
    print "------------------------------------------------------------------"

    handle = open(file_compounds)
    
    inserted_records = 0
    not_inserted = 0
    total_records = 0
    for record in Compound.parse(handle):
        total_records += 1

        ###################################################
        # start - pathway      
        for pathway in record.pathway:
            try:
                database.insert("maps_compounds", {"compoundid":str(record.entry), "mapid":str(pathway[1])})
            except:
                if debug == 1:
                    print str(record.entry), str(pathway[1]), "already inserted in modified kegg database" 
                pass
                    
            try:
                database.insert("map_names",  {"mapid":str(pathway[1]), "mapname":str(pathway[2])})
            except:
                if debug == 1:
                    print str(pathway[1]), str(pathway[2]), "already inserted in modified kegg database"
                pass
        # end - pathway
        ###################################################

        ###################################################
        # start - compounds
        if record.mass != "":
            for ion in ions.lib:

                formula_KEGG = Formula(record.formula+ion, atoms, ions)

                if len(formula_KEGG.temp_formula) == 0:
                    try:
                        database.insert("compounds", {"compoundid":record.entry, "names":str(record.name).replace("'-","-").replace("\"","'"),
                        "emp_formula":str(record.formula), "theo_mass":str(formula_KEGG.theo_mass),
                        "theo_mass_adduct":str(formula_KEGG.theo_mass_adduct), "ion":ion, "ion_mode":ion_mode})
                        inserted_records += 1
                    except:
                        not_inserted += 1
                        if debug == 1:
                            print "parsing error entry:", record.entry, ion
                else:
                    not_inserted += 1
                    if debug == 1:
                        print "parsing error entry:", record.entry, ion

        else:
            not_inserted += 1 * len(ions.lib.keys())
            if debug == 1:
                print "not inserted:", record.entry

    database.save()
    # end - compounds
    ###################################################

    database.index("compoundid_mass", "compounds", ["theo_mass_adduct"])
    database.index("compoundid_mc", "maps_compounds", ["compoundid"])

    database.execute("""INSERT INTO maps_formulae (mapid, emp_formula)
        SELECT distinct maps_compounds.mapid, compounds.emp_formula
        FROM compounds
        INNER JOIN maps_compounds WHERE
        compounds.compoundid = maps_compounds.compoundid
        ;""")
  
    print "DONE"
    print "------------------------------------------------------------------"
    print
    database.save()
    database.close()



def ReactantPairs(file_reactants, atoms, ions, ion_mode, transformation_type = 1, debug=0):

    database = SQLiteManager("MI_DB_%s" % ion_mode)
    database.create_table("reactant_pairs")
    
    print
    print "Parsing direct reactant pairs (file: %s) to MI_DB_%s.sqlite" % (file_reactants, ion_mode)
    print "------------------------------------------------------------------"

    data_records = open(file_reactants, "r")
    num_records = len(data_records.readlines())
    data_records.close()

    handle = open(file_reactants)
    perc, records_done = 0.0, 0

    pair_id = 0

    total_list = []
    for record in Map.parse(handle):

        start = []
        end = []

        for reactant in record.reactants:

            if record.reactants[reactant] < 0:
                start.append(reactant)
            else:
                end.append(reactant)

        for reactant_start in start:
            for reactant_end in end:

                cpds_start = database.select("compounds", ["compoundid", "ion", "theo_mass", "theo_mass_adduct", "emp_formula"], "WHERE compoundid = '%s'" % (reactant_start))
                cpds_end = database.select("compounds", ["compoundid", "ion", "theo_mass", "theo_mass_adduct", "emp_formula"], "WHERE compoundid = '%s'" % (reactant_end))

                if (len(cpds_start) + len(cpds_end)) == (len(ions.lib) * 2):
                        
                    pair_id += 1
                    
                    for cpd_sub in cpds_start:
                        for cpd_prod in cpds_end:

                            if cpd_sub["compoundid"] != cpd_prod["compoundid"]:


                                formula_sub = Formula(cpd_sub["emp_formula"], atoms, ions)
                                formula_prod = Formula(cpd_prod["emp_formula"], atoms, ions)

                                diff_plus = "%s+%s-%s" % (formula_sub.diff(cpd_prod["emp_formula"]), str(cpd_prod["ion"]), str(cpd_sub["ion"]))
                                diff_min = "%s+%s-%s"  % (formula_prod.diff(cpd_sub["emp_formula"]), str(cpd_sub["ion"]), str(cpd_prod["ion"]))

                                try:
                                    database.insert("reactant_pairs", {"id":pair_id, "compoundid_1":str(cpd_sub["compoundid"]),
                                    "compoundid_2":str(cpd_prod["compoundid"]), "ion_compoundid_1":str(cpd_sub["ion"]),
                                    "ion_compoundid_2":str(cpd_prod["ion"]), "type":1,
                                    "peak_pair_diff":str(cpd_prod["theo_mass"] - cpd_sub["theo_mass"]),
                                    "peak_pair_adducts_diff":str(cpd_prod["theo_mass_adduct"] - cpd_sub["theo_mass_adduct"]),
                                    "transformation":diff_min, "ion_mode":ion_mode})
                                    
                                    database.insert("reactant_pairs", {"id":pair_id, "compoundid_1":str(cpd_prod["compoundid"]),
                                    "compoundid_2":str(cpd_sub["compoundid"]), "ion_compoundid_1":str(cpd_prod["ion"]),
                                    "ion_compoundid_2":str(cpd_sub["ion"]), "type":1,
                                    "peak_pair_diff":str(cpd_sub["theo_mass"]-cpd_prod["theo_mass"]),
                                    "peak_pair_adducts_diff":str(cpd_sub["theo_mass_adduct"]-cpd_prod["theo_mass_adduct"]),
                                    "transformation":diff_plus, "ion_mode":ion_mode})
                                    
                                except sqlite3.Error, errmsg:  
                                    if debug == 1:
                                        print errmsg, pair_id, cpd_sub[0], cpd_prod[0], 1, ion_mode
                    try:
                        database.insert("reactions", {"id":pair_id, "reactionid":record.data, "type":1})

                    except sqlite3.Error, errmsg:  
                        if debug == 1:
                            print errmsg, pair_id, record.data, "d"

        records_done += 1
        if float(records_done)/num_records >= perc:
            str_perc = str(perc * 100.0) + "%"
            perc += 0.1
            database.save()
            print str_perc, 

    print
    handle.close()
    print "------------------------------------------------------------------"
    print "DONE"
    print "------------------------------------------------------------------"
    
    if transformation_type == 2:
        print
        print "Parsing indirect reactant pairs (file: %s) to MI_DB_%s.db" % (file_reactants, ion_mode)
        print "------------------------------------------------------------------"
        
        pair_id = database.max("reactant_pairs", "id", "")

        database.create_table("temp_reactant_pairs")
        tot_pairs = database.count("temp_reactant_pairs", "*", "")

        r = range(0, tot_pairs, tot_pairs/11)
        slices_sql = []
        for i in range(len(r)-1):
             slices_sql.append([r[i], tot_pairs/11])
        if r[i+1] - tot_pairs != 0:
             slices_sql.append([r[i+1], tot_pairs/11])

        perc = 0.0
                
        for slice_sql in slices_sql:

            pairs = database.select("temp_reactant_pairs", ["id1", "compoundid_1", "id2", "compoundid_2"], "LIMIT '%s', '%s';" % (slice_sql[0], slice_sql[1]))
            
            for pair in pairs:

                reactions = database.select("reactions", ["reactionid"], "WHERE id = '%s' or id = '%s';" % (pair["id1"], pair["id2"]))

                if str(pair["compoundid_1"]) != str(pair["compoundid_2"]) and len(reactions) > 1:
                    
                    pair_id += 1

                    cpds_start = database.select("compounds", ["compoundid", "ion", "theo_mass", "theo_mass_adduct", "emp_formula"], "WHERE compoundid = '%s'" % str(pair["compoundid_1"]))
                    cpds_end = database.select("compounds", ["compoundid", "ion", "theo_mass", "theo_mass_adduct", "emp_formula"], "WHERE compoundid = '%s'" % str(pair["compoundid_2"]))

                    for cpd_sub in cpds_start:
                        for cpd_prod in cpds_end:

                            if cpd_sub["compoundid"] != cpd_prod["compoundid"]:

                                formula_sub = Formula(cpd_sub["emp_formula"], atoms, ions)
                                formula_prod = Formula(cpd_prod["emp_formula"], atoms, ions)

                                diff_plus = "%s+%s-%s" % (formula_sub.diff(cpd_prod["emp_formula"]), str(cpd_prod["ion"]), str(cpd_sub["ion"]))
                                diff_min = "%s+%s-%s"  % (formula_prod.diff(cpd_sub["emp_formula"]), str(cpd_sub["ion"]), str(cpd_prod["ion"]))

                                try:
                                    database.insert("reactant_pairs", {"id":pair_id, "compoundid_1":str(cpd_sub["compoundid"]),
                                    "compoundid_2":str(cpd_prod["compoundid"]), "ion_compoundid_1":str(cpd_sub["ion"]),
                                    "ion_compoundid_2":str(cpd_prod["ion"]), "type":2,
                                    "peak_pair_diff":str(cpd_prod["theo_mass"] - cpd_sub["theo_mass"]),
                                    "peak_pair_adducts_diff":str(cpd_prod["theo_mass_adduct"] - cpd_sub["theo_mass_adduct"]),
                                    "transformation":diff_min, "ion_mode":ion_mode})
                                    
                                    database.insert("reactant_pairs", {"id":pair_id, "compoundid_1":str(cpd_prod["compoundid"]),
                                    "compoundid_2":str(cpd_sub["compoundid"]), "ion_compoundid_1":str(cpd_prod["ion"]),
                                    "ion_compoundid_2":str(cpd_sub["ion"]), "type":2,
                                    "peak_pair_diff":str(cpd_sub["theo_mass"]-cpd_prod["theo_mass"]),
                                    "peak_pair_adducts_diff":str(cpd_sub["theo_mass_adduct"]-cpd_prod["theo_mass_adduct"]),
                                    "transformation":diff_plus, "ion_mode":ion_mode})

                                except sqlite3.Error, errmsg:
                                    if debug == 1:
                                        print errmsg, pair_id, cpd_sub["compoundid"], cpd_prod["compoundid"], 2, ion_mode

                    try:
                        database.insert("reactions", {"id":pair_id, "reactionid":str(reactions[0]["reactionid"]), "type":1})
                    
                    except sqlite3.Error, errmsg:  
                        if debug == 1:
                            print errmsg, pair_id, str(reactions[0][""]), 1
                    try:
                        database.insert("reactions", {"id":pair_id, "reactionid":str(reactions[1]["reactionid"]), "type":2})

                    except sqlite3.Error, errmsg:  
                        if debug == 1:
                            print errmsg, pair_id, reactions[1][0], 2

                            
            if float(slice_sql[0]+(tot_pairs/11))/tot_pairs >= perc:
                str_perc = str(perc * 100) + "%"
                perc += 0.1
                database.save()
                print str_perc,
                
        print
        print "------------------------------------------------------------------"
        print "DONE"
        print "------------------------------------------------------------------"
        print 
    database.save()
    database.create_table("unique_transformations")

    database.index("id_r", "reactions", ["id"])
    database.index("peak_pair_adducts_diff", "unique_transformations", ["peak_pair_adducts_diff", "type", "ion_mode"])
    database.index("id_cr", "reactant_pairs", ["id"])
    database.index("compoundid_1_cr", "reactant_pairs", ["compoundid_1"])
    database.index("compoundid_2_cr", "reactant_pairs", ["compoundid_2"])
    database.index("type_cr", "reactant_pairs", ["type"])
    database.index("ion_mode_cr", "reactant_pairs", ["ion_mode"])
    database.index("compare_cr", "reactant_pairs", ["compoundid_1", "ion_compoundid_1", "transformation", "type", "ion_mode"])

    database.save()
    database.close()

"""
#####################################
# END CODE
#####################################
"""
